﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using ThinkGeo.MapSuite.Core;

namespace CacheGenerator
{
    internal class Program
    {
        /// <summary>
        /// from zoomlevel
        /// to zoomlevel
        /// bbox
        /// </summary>
        /// <param name="args"></param>
        private static void Main(string[] args)
        {
            ThreadPool.SetMaxThreads(256, 256);

            string pathFileName = @"Your MrSid file path";
            string pathName = Path.GetDirectoryName(pathFileName);

            MrSidRasterLayer rasterLayer = new MrSidRasterLayer(pathFileName);
            rasterLayer.Open();

            RectangleShape boundingBox = rasterLayer.GetBoundingBox();

            int from = 17;
            int to = 19;
            TileImageFormat imageFormat = TileImageFormat.Jpeg;
            int maxTileCountForOneThread = 100;
            GeographyUnit mapUnit = GeographyUnit.DecimalDegree;

            Collection<ZoomLevel> zoomLevels = new ZoomLevelSet().GetZoomLevels();
            Dictionary<List<BitmapTile>, TileCache> pendingTilesQueue = new Dictionary<List<BitmapTile>, TileCache>();
            for (int i = from; i < to; i++)
            {
                FileBitmapTileCache tileCache = new FileBitmapTileCache(Path.Combine(pathName, "Caches"), "DecimalDegree");
                tileCache.TileMatrix.Scale = zoomLevels[i].Scale;
                tileCache.TileMatrix.BoundingBoxUnit = mapUnit;
                tileCache.ImageFormat = imageFormat;
                tileCache.JpegQuality = 70;

                Collection<BitmapTile> tiles = tileCache.GetTiles(boundingBox);

                int currentCount = 0;
                while (currentCount < tiles.Count)
                {
                    FileBitmapTileCache newTileCache = new FileBitmapTileCache(tileCache.CacheDirectory, tileCache.CacheId);
                    newTileCache.TileMatrix.Scale = zoomLevels[i].Scale;
                    newTileCache.TileMatrix.BoundingBoxUnit = mapUnit;
                    newTileCache.ImageFormat = imageFormat;
                    newTileCache.JpegQuality = 70;

                    List<BitmapTile> pendingTiles = tiles.Skip(currentCount).Take(maxTileCountForOneThread).ToList();
                    currentCount += pendingTiles.Count;
                    pendingTilesQueue.Add(pendingTiles, newTileCache);
                }
            }

            int totalCount = pendingTilesQueue.SelectMany(t => t.Key).Count();
            int processCount = 0;
            Parallel.ForEach(pendingTilesQueue, drawingTiles =>
            {
                MrSidRasterLayer drawingLayer = new MrSidRasterLayer(pathFileName);
                drawingLayer.Open();
                Collection<SimpleCandidate> candidates = new Collection<SimpleCandidate>();
                foreach (var tile in drawingTiles.Key)
                {
                    Bitmap bitmap = new Bitmap(drawingTiles.Value.TileMatrix.TileWidth, drawingTiles.Value.TileMatrix.TileHeight);
                    GdiPlusGeoCanvas geoCanvas = new GdiPlusGeoCanvas();
                    geoCanvas.BeginDrawing(bitmap, tile.BoundingBox, drawingTiles.Value.TileMatrix.BoundingBoxUnit);
                    drawingLayer.Draw(geoCanvas, candidates);
                    geoCanvas.EndDrawing();

                    tile.Bitmap = bitmap;
                    drawingTiles.Value.SaveTile(tile);
                    tile.Dispose();
                    bitmap.Dispose();

                    Interlocked.Increment(ref processCount);
                    int progress = processCount * 100 / totalCount;
                    if (processCount % 100 == 0)
                    {
                        Console.Write("Progress {0} % ({1} / {2})\r", progress, processCount, totalCount);
                    }
                }
                rasterLayer.Close();
            });

            Console.Write("Finished.");
        }
    }
}